from pathlib import Path

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
ROOT = Path.home()
DATA_DIR = Path.home() / ".local" / "orcbench"
MODEL_DIR = DATA_DIR / "models"
MODELS = 25
URL = "https://rcs.uwaterloo.ca/~ryan/files/models"


# Model Creation Globals
RAW_DATA_URL = "https://azurecloudpublicdataset2.blob.core.windows.net/azurepublicdatasetv2/azurefunctions_dataset2019/azurefunctions-dataset2019.tar.xz"

RAW_DATA_DIR = DATA_DIR / "raw_data"
POISSON_SEED = 1337
NUM_DATA_FILES = 40
DEFAULT_DAYS = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]

STARTTIME = 720
CAPTURE_WINDOW = 30
INVOCATION_CUTOFF = 10


# Logger
FORMAT = "[%(asctime)s] %(message)s"



import math
from time import time

import numpy as np
import scipy.stats as st
from scipy.interpolate import interp1d
from sklearn.cluster import KMeans

# from statsmodels.distributions.empirical_distribution import ECDF

length = 2000
qseq = np.linspace(1e-6, 1 - (1e-6), length)
# Taken from Seaborns implementation
class ECDF:
    """Univariate empirical cumulative distribution estimator."""

    def __init__(self, x):
        """Inner function for ECDF of one variable."""
        self.x_max = max(x)
        self.x_min = min(x)
        sorter = x.argsort()
        x = x[sorter]
        weights = np.ones_like(x)
        weights = weights[sorter]
        y = weights.cumsum()

        self.x = np.r_[0, x]
        self.y = np.r_[0, y]
        self.y = self.y / self.y.max()
        self.inter = interp1d(self.x, self.y, fill_value=0.0)

    def __call__(self, x, weights=None):
        if x > self.x_max:
            return 1.0
        if x < self.x_min:
            return 0.0

        return self.inter(x)


class ECDFWrapper:
    def __init__(self, k, inverse):

        # Filter out non-sensical, this is some a bug earlier in the
        # pipeline
        k = np.asarray(k)
        filter = k < 10e7
        k = k[filter]
        filter = k > (1.0 / 1440)
        k = k[filter]

        self.data = np.asarray(k, dtype="float32")
        self.ecdf = ECDF(self.data)

        self.x_min = min(self.data)
        self.x_max = max(self.data)
        # Dont go to negative inf X

        if inverse:
            self.inverse = quantile(self, qseq)
            assert not np.isnan(self.inverse).any()
        else:
            self.inverse = None

    def _re_init(self, length=1000):
        qseq = np.linspace(1e-6, 1 - (1e-6), length)
        print(self.x_min, self.x_max)
        self.inverse = quantile(self, qseq)

    def __call__(self, x):
        return self.ecdf(x)

    def data(self):
        return self.data

    def sample(self, num=1, random_state=int(time())):
        np.random.seed(random_state)
        samples = []
        # TODO: Fix bug with lowest percentile
        for _ in range(0, num):
            p = np.random.randint(0, high=length - 1)
            samples.append(1 / self.inverse[p])
        return samples


def ecdf(keys, inverse=True):
    return ECDFWrapper(keys, inverse)


def emd(m, n):
    return st.wasserstein_distance(m, n)


def quantile(cdf, qs):
    return [_quantile_help(cdf, q) for q in qs]


def _quantile_help(cdf, q):
    try:
        l = math.log10(min(cdf.data))
        h = math.log10(max(cdf.data))
    except:
        for x in cdf.data:
            if x > 10e5:
                print(x)
            if x < 0:
                print(x)
        print(cdf.data.size)
        print(cdf.x_min, cdf.x_max)
        exit(0)
    qseq = np.logspace(l, h, 100000)
    # qseq = np.linspace(cdf.x_min, cdf.x_max, 1000000)
    lo = 0
    hi = len(qseq) - 1
    while lo < hi:
        mid = (lo + hi) // 2
        v = qseq[mid]
        if cdf(v) < q:
            lo = mid + 1
        else:
            hi = mid

    lo = qseq[lo]

    return lo


def epmeans_cluster(clist, k=2):
    myn = len(clist)
    print(myn)
    qmat = []
    for i in range(0, myn):
        if np.isnan(clist[i].inverse).any():
            continue
        qmat.append(clist[i].inverse)
    print("Clustering")
    cluster = KMeans(n_clusters=k, init="k-means++")
    cluster.fit(qmat)

    means = cluster.cluster_centers_
    list2 = []
    list1 = []

    print("Done Clustering")
    for i in range(0, myn):
        list1.append(ecdf(qmat[i], inverse=False))

    for i in range(0, k):
        list2.append(ecdf(means[i]))

    distance_matrix = []
    for i, m in enumerate(list1):
        tmp = []
        for _, n in enumerate(list2):
            tmp.append(emd(m.data, n.data))
        distance_matrix.append(tmp)

    labels = np.argmin(distance_matrix, axis=1)

    # Get the total EMD for the clusters and their associated models attached
    # to it
    s = 0
    for i, l in enumerate(labels):
        s += distance_matrix[i][l]

    return (list2, labels, s)



import random
import math
import numpy as np

from .orcglobals import CAPTURE_WINDOW

MAX_ROUND = 7


def inverse_poisson(start_min, mu):
    if mu == 0:
        return np.asarray([])
    interarrivals = np.random.exponential(1.0 / mu, mu)
    interarrivals = np.around(interarrivals, decimals=MAX_ROUND)
    interarrivals = np.cumsum(interarrivals)
    interarrivals = interarrivals + start_min
    return interarrivals


def resample_as_poisson(t, begin, end):
    my_values = []
    if sum(t[begin:end]) == 0:
        return np.asarray([end - begin])

    for min, val in enumerate(t[begin:end]):
        l = inverse_poisson(min, val)
        my_values.extend(l)

    my_values = np.asarray(my_values)
    my_values.sort()

    if begin == 0:
        prev = 0
    prev = 0
    i = begin
    while (i > 0) and (t[i - 1]) == 0:
        i -= 1
        prev -= 1

    final_values = [my_values[0] - prev]
    # Now go through this new interarrival timeline and calculate IA rate
    prev = my_values[0]
    for i in range(1, len(my_values)):
        next_value = my_values[i]

        # Take what the last IA value was so we double its weight
        if next_value == prev:
            next_value = prev + (1.0 / math.pow(10, -1 * MAX_ROUND))
        rounded = next_value - prev

        final_values.append(rounded)
        prev = next_value

    return np.asarray(final_values)


def calc_ia_time(func):
    t = func["timeline"]
    # Sunday + 12Hrs
    chunks = np.arange(1440 + 720, (5 * 1440), 1440)
    final_list = []
    for x in chunks:
        resampled = resample_as_poisson(t, x, x + CAPTURE_WINDOW)
        final_list.extend(resampled)

    func["ia"] = np.asarray(final_list, dtype="float32")


def calc_invocs_list(trace, run_for, unit=1):
    if run_for % unit == 0:
        invocs_list = [0] * int((run_for) / unit)
    else:
        invocs_list = [0] * int((run_for + unit) / unit)
    cur = 0
    invocs = 0
    ii = 0
    for t in trace:
        v = t[0]
        f = 1
        if int(v) >= run_for:
            break
        if int(v) > cur + unit:
            if ii == 0:
                i = 0
            else:
                i = int(trace[ii - 1][0] / unit)
            invocs_list[i] = invocs
            invocs = 0
            cur += unit
        if int(v) == cur + unit:
            if ii == 0:
                i = 0
            else:
                i = int(trace[ii - 1][0] / unit)
            invocs_list[i] = invocs
            invocs = f
            cur += unit
        else:
            invocs += f
        ii += 1

    ii -= 1
    if len(trace):
        invocs_list[int(trace[ii][0] / unit)] = invocs
    return invocs_list


def calculate_ia_freq(t, begin, end):
    return resample_as_poisson(t, begin, end)



import os
import json
import uuid
from pathlib import Path

import numpy as np

from .orcglobals import MODEL_DIR

loaded_models = []


class Job:
    def __init__(self, model, num_functions, ids=None):
        assert num_functions != 0
        self.model = model
        self.f = num_functions
        self.ids = ids
        self.rand = np.random.default_rng(np.random.randint(0, 2 ** 32))

        if ids is None:
            self.ids = [uuid.uuid1().hex for _ in range(0, num_functions)]

        self.run_times = {id: self.model.sample_cpu() for id in self.ids}
        self.memory = {id: self.model.sample_mbs() for id in self.ids}

    def run_trace(self, run_for):
        cur_time = 0
        # noise = model.noise()
        trace = []
        # Lets lookst at if we create an ecdf from scratch by looking at the
        # upper and lower bounds of the ecdf of each cluster.
        while cur_time < run_for:
            samples = np.array(self.model.sample(10000))
            cur_time += samples[0]
            i = 1
            while i < len(samples):
                time = samples[i]
                i += 1
                if np.isinf(time) or time < 0:
                    continue

                events = self.f
                ids = self.ids.copy()
                self.rand.shuffle(ids)
                for v in sorted(self.rand.uniform(cur_time, cur_time + time, events)):
                    id = ids[0]
                    ids.pop(0)
                    trace.append((v, id, self.model.name))
                #                    trace.append((v, id, self.model.name,
                #                        self.run_times[id], self.memory[id]))

                cur_time += time

                if cur_time >= run_for:
                    break
        return [t for t in trace if t[0] < run_for]


def load_models():
    models_dir = Path(__file__).parent.parent / "models";
    for f in os.listdir(models_dir):
        model_path = models_dir.absolute() / f
        with open(model_path, "rb") as file:
            loaded_models.append(Model(f, json.load(file)))

    print("Model Functions: ", sum([x.weight for x in loaded_models]))

    return loaded_models


def _sample(inverse_ecdf, num, rand):
    samples = []
    for _ in range(0, num):
        p = rand.integers(0, high=len(inverse_ecdf) - 1)
        samples.append(1 / inverse_ecdf[p])
    return samples


def python_function_template(runtime):
    program = """
    import time
    import mmap
    def lambda():
        t = time.time()
        # Busy wait so we consume cpu
        while (time.time() - t) < {0}:
            pass
    """.format(
        runtime
    )

    return program


def create_function(model, language="python"):
    mbs = model.sample_mbs()
    runtime = model.sample_cpu()

    if language == "python":
        program = python_function_template(runtime)
    else:
        program = ""

    return program


class Model:
    def __init__(self, model_name, model_dict):
        self.name = model_name
        self.inverse_invocation_ecdf = model_dict["inverse_ecdf"]
        self.inverse_memory_ecdf = model_dict["mbs"]
        self.inverse_cpu_ecdf = model_dict["cpu"]
        self.weight = model_dict["num_functions"]
        self.rand = np.random.default_rng(np.random.randint(0, 2 ** 32))

    def sample(self, num=1):
        return _sample(self.inverse_invocation_ecdf, num, self.rand)

    def sample_mbs(self, num=1):
        return _sample(self.inverse_memory_ecdf, num, self.rand)

    def sample_cpu(self, num=1):
        return _sample(self.inverse_cpu_ecdf, num, self.rand)


def get_jobs(n, scale, seed):
    np.random.seed(seed)

    models = load_models()

    job_map = {}
    for m in models:
        jobs = []
        num_functions = int(m.weight * scale)
        if num_functions == 0:
            continue
        groups = int(num_functions / n)
        extra = num_functions % n
        j = [Job(m, n) for _ in range(0, groups)]
        jobs.extend(j)
        if extra != 0:
            jobs.append(Job(m, extra))
        job_map[m.name] = jobs

    return job_map



import requests
import click
import json
import subprocess
import sys
import tempfile
import tarfile
import pickle
import logging
import os
import collections

from itertools import groupby
from operator import itemgetter
from pathlib import Path

import pandas as pd
import numpy as np

import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt

# We do this cause the models file can be directly called to run
sys.path.append(str(Path(__file__).parent.absolute()))

from internals.epmeans import ecdf, epmeans_cluster
from internals.stats import calc_ia_time

from internals.orcglobals import RAW_DATA_URL, RAW_DATA_DIR, NUM_DATA_FILES
from internals.orcglobals import FORMAT, DEFAULT_DAYS, CAPTURE_WINDOW
from internals.orcglobals import INVOCATION_CUTOFF, MODEL_DIR
from internals.stats import calc_invocs_list

from internals.run import get_jobs

raw_data_dir = RAW_DATA_DIR

logging.basicConfig(format=FORMAT, level=os.environ.get("LOGLEVEL", "INFO"))
log = logging.getLogger(__name__)


def graph(group, ax=None):
    fig = None
    if ax is None:
        fig, ax = plt.subplots()

    df = group["center"]
    log.info("Graphing CDFS of models")
    for m in group["functions"]:
        e = m["ecdf"]
        ax.plot(e.ecdf.x, e.ecdf.y, alpha=0.3)

    ax.plot(df.ecdf.x, df.ecdf.y, label="Centroid", color="black")
    ax.set_xscale("log")

    vals = []

    # Rough evaluation
    for c in group["functions"]:
        vals.append(np.mean(c["sub"]))
    log.info("Real ", np.mean(vals))
    s = group["center"].sample(100000)
    cm = np.cumsum(s)
    ma = int(max(cm) + 1)
    ra = np.arange(ma + 1, dtype="float32")
    bins, _ = np.histogram(cm, bins=ra)
    log.info("Model", np.mean(bins))

    return fig


def parse_data(days=DEFAULT_DAYS):
    # Tuples of prefixes for files as well as what we want to group the data on
    days_full = ["{}".format(str(x).zfill(2)) for x in range(1, 13)]
    days = ["{}".format(str(x).zfill(2)) for x in days]
    prefixes = [
        ("app_memory_percentiles.anon.d", "HashApp", days_full),
        ("function_durations_percentiles.anon.d", "HashFunction", days_full),
        ("invocations_per_function_md.anon.d", "HashFunction", days),
    ]

    # Grab all the files for specific data and read it in
    g = []
    total = (len(days_full) * 2) + len(days)
    timers_count = 0
    with click.progressbar(length=total, label="Loading Data") as bar:
        for pre in prefixes:
            files = [f"{pre[0]}{d}.csv" for d in pre[2]]
            invoc = []
            for i, f in enumerate(files):
                df = pd.read_csv(os.path.join(raw_data_dir, f))
                df["days"] = [i + 1] * len(df.index)
                # Filter away our timer triggers
                before = len(df)
                try:
                    df = df[df["Trigger"] != "timer"]
                except:
                    pass
                after = len(df)
                timers_count += before - after
                invoc.append(df)
                bar.update(1)

            con = pd.concat(invoc, axis=0, ignore_index=True)
            g.append(con.groupby(pre[1]))

    return g[0], g[1], g[2]


def download_data():
    global raw_data_dir
    if len(os.listdir(raw_data_dir)) == NUM_DATA_FILES:
        log.info("Data already downloaded and extracted, continuing..")
        return

    log.info("Downloading data")
    response = requests.get(RAW_DATA_URL)
    f = tempfile.NamedTemporaryFile()
    f.write(response.content)
    log.info("Extracting data")
    with tarfile.open(f.name, mode="r:xz") as tf:
        tf.extractall(path=raw_data_dir)
    f.close()


def aggregate_data():
    app, durations, invocations = parse_data()
    # Set up the invocation
    duration_keys = set([name for name, _ in durations])
    invocation_keys = set([name for name, _ in invocations])
    common = list(duration_keys.intersection(invocation_keys))
    log.info(f"{len(duration_keys)}, {len(invocation_keys)}, {len(common)}")
    functions = []

    exception_cont = 0
    with click.progressbar(length=len(common), label="Parsing Data") as bar:
        for group in common:
            bar.update(1)
            # Get timelines of each application.
            trigger = None
            hashApp = None
            hashFunction = None
            hashOwner = None
            t = [[0] * 1440] * len(DEFAULT_DAYS)
            for row in invocations.get_group(group).to_numpy().tolist():
                t[row[-1] - 1] = np.asarray(list(map(int, row[4:-1])), dtype="uint32")
                hashOwner = row[0]
                hashApp = row[1]
                hashFunction = row[2]
                trigger = row[3]

            t = np.asarray([i for sl in t for i in sl], dtype="uint32")

            # Aggregate data into one object
            d = dict(
                timeline=t,
                hash=hashFunction,
                owner=hashOwner,
                app=hashApp,
                trigger=trigger,
            )

            percentiles = [1, 25, 50, 75, 99, 100]
            percentiles_cpu = ["percentile_Average_{}".format(x) for x in percentiles]
            percentiles_ram = [
                "AverageAllocatedMb_pct{}".format(x) for x in percentiles
            ]
            retrieve_cpu = ["Average", "Minimum", "Maximum"] + percentiles_cpu
            retrieve_ram = ["AverageAllocatedMb"] + percentiles_ram
            cpu = []

            # Attach CPU and Memory data to the function
            try:
                for _, row in durations.get_group(group).iterrows():
                    cpu.append(np.array([row[x] for x in retrieve_cpu]))
                s = len(cpu)
                cpu = np.sum(cpu, axis=0) / s
                mem = []

                for _, row in app.get_group(hashApp).iterrows():
                    mem.append(np.array([row[x] for x in retrieve_ram]))
                s = len(mem)
                mem = np.sum(mem, axis=0) / s

                for ii, x in enumerate(retrieve_cpu):
                    d[x] = cpu[ii]
                for ii, x in enumerate(retrieve_ram):
                    d[x] = mem[ii]

            # If function does not have an associated app then discard it
            except:
                exception_cont += 1
                continue

            functions.append(d)

    return functions


def filter_and_ia(functions):
    for func in functions:
        calc_ia_time(func)

    return functions


def invert_and_ecdf(functions):
    # Cluster on our Frequency Data
    for func in functions:
        vals = ecdf(np.reciprocal(func["ia"], dtype="float32"))
        func["ecdf"] = vals
        del func["ia"]

    return functions


snapshot_one = raw_data_dir / "snapshot_one.pickle"
snapshot_two = raw_data_dir / "snapshot_two.pickle"
snapshot_three = raw_data_dir / "snapshot_three.pickle"
snapshot_three_b = raw_data_dir / "snapshot_three_b.pickle"
snapshot_four = raw_data_dir / "snapshot_four.pickle"


def s1(functions):
    download_data()
    functions = aggregate_data()
    with open(snapshot_one, "wb") as f:
        pickle.dump(functions, f)

    return functions


def split_and_collect(final, command, NUM_PER=200):
    global raw_data_dir
    funcs = [final[i : i + NUM_PER] for i in range(0, len(final), NUM_PER)]
    path = raw_data_dir / "tmp"
    path.mkdir(parents=True, exist_ok=True)
    this_file = os.path.abspath(__file__)
    procs = []
    files = []
    for i, f in enumerate(funcs):
        p = path / f"{i}.pickle"
        files.append(p)
        with open(p, "wb") as file:
            pickle.dump(f, file)
        log.info(p)
        os.sync()

        cmd = f"{sys.executable} {this_file} {command} {p}"
        proc = subprocess.Popen(
            cmd,
            shell=True,
            stdin=None,
            stdout=sys.stdout,
            stderr=sys.stderr,
            close_fds=True,
        )
        procs.append(proc)

    for i, p in enumerate(procs):
        p.wait()
        log.info(files[i])

    functions = []
    for f in files:
        with open(f, "rb") as file:
            functions.extend(pickle.load(file))

    for f in files:
        os.unlink(f)

    return functions


def s2(functions):
    log.info("Filtering and calculating interarrival times")
    final = []
    # Apply minimum invocation filter
    for func in functions:
        chunks = np.arange(720, len(DEFAULT_DAYS) * 1440, 1440)
        overall = []
        check = []
        for x in chunks:
            t = func["timeline"]
            overall.append(np.sum(t[x : x + CAPTURE_WINDOW]))
            check.extend(t[x : x + CAPTURE_WINDOW])

        # If there doesnt exist a window of time in any of the
        # days we are looking at the has at least INVOCATION_CUTOFF
        # invocations then we don't use it
        if not any([i > INVOCATION_CUTOFF for i in overall]):
            continue

        func["sub"] = check
        final.append(func)

    functions = split_and_collect(final, "ia")
    del final

    with open(snapshot_two, "wb") as f:
        pickle.dump(functions, f)

    return functions


def s3(functions):
    log.info("Calculate top 1%")
    weights = []
    for func in functions:
        weights.append(func["ia"].size)
    ninenine = np.percentile(weights, 99)

    functions_top = [func for func in functions if func["ia"].size >= ninenine]

    functions_bottom = [func for func in functions if func["ia"].size < ninenine]

    functions_bottom = split_and_collect(functions_bottom, "ecdf")
    functions_top = split_and_collect(functions_top, "ecdf")
    with open(snapshot_three, "wb") as f:
        pickle.dump((functions_top, functions_bottom), f)


def group_up(functions, centers, labels):
    log.info(collections.Counter(labels))
    log.info(len(centers))
    zipped = list(zip(functions, labels))
    zipped.sort(key=itemgetter(1))
    group = groupby(zipped, itemgetter(1))
    groups = []
    for key, data in group:
        d = {"center": centers[key], "functions": [d[0] for d in data]}
        groups.append(d)

    return groups


def s4(functions):
    log.info("Use EP-Means Clustering")
    functions_top = functions[0]
    functions_bottom = functions[1]

    functions_bottom = [f for f in functions_bottom if f["ecdf"] is not None]
    functions_top = [f for f in functions_top if f["ecdf"] is not None]

    data = [f["ecdf"] for f in functions_bottom]
    log.info("Use EP-Means Clustering low")
    centers, labels, error = epmeans_cluster(data, k=20)
    group_bottom = group_up(functions_bottom, centers, labels)

    data = [f["ecdf"] for f in functions_top]
    log.info("Use EP-Means Clustering high")
    centers, labels, error = epmeans_cluster(data, k=5)
    group_top = group_up(functions_top, centers, labels)

    group_bottom.extend(group_top)

    log.info("Snapshotting")
    with open(snapshot_four, "wb") as f:
        pickle.dump(group_bottom, f)

    return group_bottom


def create_models(data):
    global raw_data_dir
    global snapshot_one
    global snapshot_two
    global snapshot_three
    global snapshot_three_b
    global snapshot_four

    raw_data_dir = Path(data)
    snapshot_one = raw_data_dir / "snapshot_one.pickle"
    snapshot_two = raw_data_dir / "snapshot_two.pickle"
    snapshot_three = raw_data_dir / "snapshot_three.pickle"
    snapshot_four = raw_data_dir / "snapshot_four.pickle"

    functions = None
    retrieve_map = {
        snapshot_one: s1,
        snapshot_two: s2,
        snapshot_three: s3,
        snapshot_four: s4,
    }

    snaps = list(retrieve_map.keys())
    vals = list(retrieve_map.values())
    for cur in range(0, len(snaps)):
        if snaps[cur].exists():
            log.info(f"{snaps[cur]} exists")
            continue
        else:
            if (cur - 1) != -1:
                with open(snaps[cur - 1], "rb") as f:
                    functions = pickle.load(f)

            functions = vals[cur](functions)

    if functions == None:
        with open(snaps[-1], "rb") as f:
            functions = pickle.load(f)

    log.info("Graphing Models")
    models_dir = Path("graphs")
    models_dir.mkdir(parents=True, exist_ok=True)
    for i, group in enumerate(functions):
        fig = graph(group)
        fig.savefig(models_dir / f"{i}-model.svg")
        plt.close(fig)

    log.info("Create Models")
    for i, model in enumerate(functions):
        log.info("Creating - ", i)
        mbs = [func["AverageAllocatedMb"] for func in model["functions"]]
        cpu = [func["Average"] for func in model["functions"]]
        model["mbs"] = ecdf(mbs).inverse
        model["cpu"] = ecdf(cpu).inverse

    models_dir = MODEL_DIR
    models_dir.mkdir(parents=True, exist_ok=True)
    st = 720 + 2880
    end = st + CAPTURE_WINDOW
    evaluation = {}
    for i, group in enumerate(functions):
        k = f"{i}-model"
        p = models_dir / k
        model = group["center"]
        with open(p, "w") as f:
            m = dict()
            m["inverse_ecdf"] = model.inverse
            m["num_functions"] = len(group["functions"])
            m["mbs"] = group["mbs"]
            m["cpu"] = group["cpu"]
            json.dump(m, f)

        arr = np.zeros(30)
        for f in group["functions"]:
            t = np.asarray(f["timeline"][st:end])
            arr = np.add(arr, t)
        evaluation[k] = arr

    models_dir = Path("evaluation")
    models_dir.mkdir(parents=True, exist_ok=True)
    job_map = get_jobs(100, 1.0, 0)
    if job_map is None:
        log.error("Could not get models - are the models there?")
        exit(0)

    xs = np.arange(0, 30, 1)
    for name, jobs in job_map.items():
        p = models_dir / name
        trace = evaluation[name]
        arr = np.zeros(30)

        for j in jobs:
            t = np.asarray(calc_invocs_list(j.run_trace(30), 30))
            arr = np.add(arr, t)
        fig, ax = plt.subplots()

        ax.plot(xs, trace, label="Trace")
        ax.plot(xs, arr, label="Model")
        ax.legend()
        fig.savefig(p)


if __name__ == "__main__":
    if sys.argv[1] == "ecdf":
        with open(sys.argv[2], "rb") as f:
            functions = pickle.load(f)
            functions = invert_and_ecdf(functions)

        with open(sys.argv[2], "wb") as f:
            pickle.dump(functions, f)

    elif sys.argv[1] == "ia":
        with open(sys.argv[2], "rb") as f:
            functions = pickle.load(f)
            functions = filter_and_ia(functions)

        with open(sys.argv[2], "wb") as f:
            pickle.dump(functions, f)



"""
OrcBench is a benchmark for serverless. OrcBench produces traces 
which can be used by serveless platforms to test their service
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⣿⣿⡄⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⣀⣤⣶⣧⣄⣉⣉⣠⣼⣶⣤⣀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⢰⣿⣿⣿⣿⡿⣿⣿⣿⣿⢿⣿⣿⣿⣿⡆⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⣼⣤⣤⣈⠙⠳⢄⣉⣋⡡⠞⠋⣁⣤⣤⣧⠀⠀⠀⠀⠀⠀⠀
⠀⢲⣶⣤⣄⡀⢀⣿⣄⠙⠿⣿⣦⣤⡿⢿⣤⣴⣿⠿⠋⣠⣿⠀⢀⣠⣤⣶⡖⠀
⠀⠀⠙⣿⠛⠇⢸⣿⣿⡟⠀⡄⢉⠉⢀⡀⠉⡉⢠⠀⢻⣿⣿⡇⠸⠛⣿⠋⠀⠀
⠀⠀⠀⠘⣷⠀⢸⡏⠻⣿⣤⣤⠂⣠⣿⣿⣄⠑⣤⣤⣿⠟⢹⡇⠀⣾⠃⠀⠀⠀
⠀⠀⠀⠀⠘⠀⢸⣿⡀⢀⠙⠻⢦⣌⣉⣉⣡⡴⠟⠋⡀⢀⣿⡇⠀⠃⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⢸⣿⣧⠈⠛⠂⠀⠉⠛⠛⠉⠀⠐⠛⠁⣼⣿⡇⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠸⣏⠀⣤⡶⠖⠛⠋⠉⠉⠙⠛⠲⢶⣤⠀⣹⠇⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⢹⣿⣶⣿⣿⣿⣿⣿⣿⣶⣿⡏⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⠉⠉⠉⠛⠛⠛⠛⠉⠉⠉⠁⠀⠀⠀⠀⠀⠀⠀⠀⠀
WAAAGH!
"""

import click
import secrets

from .internals import run
from .internals.orcglobals import MODELS, CONTEXT_SETTINGS, RAW_DATA_DIR

@click.group()
def cli():
    pass


@cli.command()
@click.option("--N", default=10, help="Sampling Group Size")
@click.option("--scale", default=0.25, help="Percentage of workload")
@click.option("--seed", default=secrets.randbits(32), help="Seed for sampling")
@click.option("--runtime", default=30, help="How long should trace run for")
@click.option("--out", default="trace.out", help="Output file for the trace")
def trace(n, scale, seed, runtime, out):
    job_map = run.get_jobs(n, scale, seed)
    trace = []
    total = 0
    for _, jobs in job_map.items():
        total += len(jobs)
    with click.progressbar(length=total, label="Running Trace") as bar:
        for _, jobs in job_map.items():
            for job in jobs:
                t = job.run_trace(runtime)
                trace.extend(t)
                bar.update(1)
    trace.sort(key=lambda x: x[0])
    click.echo(f"Total invocations: {len(trace)}")
    ids = set()
    with open(out, "w") as f:
        with click.progressbar(length=len(trace), label="Writing trace") as bar:
            for t in trace:
                ids.add(t[1])
                f.write(f"{t[0]}, {t[1]}, {t[2]}\n")
                bar.update(1)
    click.echo(f"Unique Functions: {len(ids)}")

@cli.command()
@click.argument("model")
@click.option("--language", default="python")
def function(model, language):
    loaded_models = run.load_models()
    found = None
    for m in loaded_models:
        if m.name == model:
            found = model
    program_str = run.create_function(found, language=language)
    click.echo(program_str)


@cli.command()
@click.option("--data", default=RAW_DATA_DIR, type=click.Path(exists=True))
def create(data):
    from .models import create_models
    create_models(data)


if __name__ == "__main__":
    main_cli = click.CommandCollection(sources=[cli], context_settings=CONTEXT_SETTINGS)
    main_cli(prog_name="OrcBench")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orcbench', 'orcbench.internals']

package_data = \
{'': ['*'], 'orcbench': ['models/*']}

install_requires = \
['click>=8.1.3,<9.0.0', 'numpy>=1.23.1,<2.0.0']

setup_kwargs = {
    'name': 'orcbench',
    'version': '0.1.2',
    'description': 'OrcBench: A Representative Serverless Benchmark',
    'long_description': "\n    ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⣿⣿⡄⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠀⠀⠀⣀⣤⣶⣧⣄⣉⣉⣠⣼⣶⣤⣀⠀⠀⠀⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠀⢰⣿⣿⣿⣿⡿⣿⣿⣿⣿⢿⣿⣿⣿⣿⡆⠀⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠀⣼⣤⣤⣈⠙⠳⢄⣉⣋⡡⠞⠋⣁⣤⣤⣧⠀⠀⠀⠀⠀⠀⠀\n    ⠀⢲⣶⣤⣄⡀⢀⣿⣄⠙⠿⣿⣦⣤⡿⢿⣤⣴⣿⠿⠋⣠⣿⠀⢀⣠⣤⣶⡖⠀\n    ⠀⠀⠙⣿⠛⠇⢸⣿⣿⡟⠀⡄⢉⠉⢀⡀⠉⡉⢠⠀⢻⣿⣿⡇⠸⠛⣿⠋⠀⠀\n    ⠀⠀⠀⠘⣷⠀⢸⡏⠻⣿⣤⣤⠂⣠⣿⣿⣄⠑⣤⣤⣿⠟⢹⡇⠀⣾⠃⠀⠀⠀\n    ⠀⠀⠀⠀⠘⠀⢸⣿⡀⢀⠙⠻⢦⣌⣉⣉⣡⡴⠟⠋⡀⢀⣿⡇⠀⠃⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⢸⣿⣧⠈⠛⠂⠀⠉⠛⠛⠉⠀⠐⠛⠁⣼⣿⡇⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠸⣏⠀⣤⡶⠖⠛⠋⠉⠉⠙⠛⠲⢶⣤⠀⣹⠇⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠀⠀⠀⢹⣿⣶⣿⣿⣿⣿⣿⣿⣶⣿⡏⠀⠀⠀⠀⠀⠀⠀⠀⠀\n    ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⠉⠉⠉⠛⠛⠛⠛⠉⠉⠉⠁⠀⠀⠀⠀⠀⠀⠀⠀⠀\n## OrcBench: A Representative Serverless Benchmark\nOrcBench generates a workload trace which can be consumed by serverless\nplatforms to test their service. The models used by OrcBench are modeled off of\nMicrosoft Azure and whos creation is outlined in our paper (referenced below).\n\n### Installation\n```\npip install orcbench\n```\n\n### Quick Start\n```\norcbench trace\n```\nThis will produce a standard trace (25% that of the original Microsoft\nWorkload, `--scale 0.25`). nd produces jobs which will sample for 10 models\n(`--N 10`). With a runtime (`--runtime 30`) of 30 minutes. The seed (`--seed`)\ncan be optionally set to produce deterministic workloads. The outputted\n(`--out`) trace is by default sent to `trace.out`. \n\n## Referencing Us \nIf you use this benchmark please make sure to cite us using the following citation\n\nPDF of the paper can be found - [OrcBench: A Representative Serverless Benchmark](https://rcs.uwaterloo.ca/~ryan/files/orcbench.pdf)\n\n```\n@inproceedings{hancock:benchmarking,\n  author    = {Ryan Hancock and\n               Sreeharsha Udayashankar and\n               Ali José Mashtizadeh and\n               Samer Al-Kiswany},\n  title     = {OrcBench: A Representative Serverless Benchmark},\n  booktitle = {Proceedings of the15th International Conference on Cloud Computing (CLOUD'22)}\n  publisher = {{IEEE}},\n  year      = {2022},\n  doi       = {10.1109/CLOUD55607.2022.00028},\n}\n```\n\n",
    'author': 'Ryan Hancock',
    'author_email': 'krhancoc@uwaterloo.ca',
    'maintainer': 'Ryan Hancock',
    'maintainer_email': 'krhancoc@uwaterloo.ca',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
